@@ -8,7 +8,7 @@
 
 ! Fortran 2018 Clause 17
 
-include '../include/flang/Runtime/magic-numbers.h'
+#include '../include/flang/Runtime/magic-numbers.h'
 
 module ieee_arithmetic
   ! F18 Clause 17.1p1:
@@ -18,13 +18,20 @@ module ieee_arithmetic
   use __fortran_ieee_exceptions
 
   use __fortran_builtins, only: &
+    ieee_away => __builtin_ieee_away, &
+    ieee_down => __builtin_ieee_down, &
     ieee_fma => __builtin_fma, &
+    ieee_int => __builtin_ieee_int, &
     ieee_is_nan => __builtin_ieee_is_nan, &
     ieee_is_negative => __builtin_ieee_is_negative, &
     ieee_is_normal => __builtin_ieee_is_normal, &
+    ieee_nearest => __builtin_ieee_nearest, &
     ieee_next_after => __builtin_ieee_next_after, &
     ieee_next_down => __builtin_ieee_next_down, &
     ieee_next_up => __builtin_ieee_next_up, &
+    ieee_other => __builtin_ieee_other, &
+    ieee_real => __builtin_ieee_real, &
+    ieee_round_type => __builtin_ieee_round_type, &
     ieee_scalb => scale, &
     ieee_selected_real_kind => __builtin_ieee_selected_real_kind, &
     ieee_support_datatype => __builtin_ieee_support_datatype, &
@@ -33,19 +40,81 @@ module ieee_arithmetic
     ieee_support_inf => __builtin_ieee_support_inf, &
     ieee_support_io => __builtin_ieee_support_io, &
     ieee_support_nan => __builtin_ieee_support_nan, &
+    ieee_support_rounding => __builtin_ieee_support_rounding, &
     ieee_support_sqrt => __builtin_ieee_support_sqrt, &
     ieee_support_standard => __builtin_ieee_support_standard, &
     ieee_support_subnormal => __builtin_ieee_support_subnormal, &
-    ieee_support_underflow_control => __builtin_ieee_support_underflow_control
+    ieee_support_underflow_control => __builtin_ieee_support_underflow_control, &
+    ieee_to_zero => __builtin_ieee_to_zero, &
+    ieee_up => __builtin_ieee_up
+
 
   implicit none
 
-  type :: ieee_class_type
+  ! Set PRIVATE by default to explicitly only export what is meant
+  ! to be exported by this MODULE.
+  private
+
+  ! Explicitly export the symbols from __fortran_builtins
+  public :: ieee_away
+  public :: ieee_down
+  public :: ieee_fma
+  public :: ieee_int
+  public :: ieee_is_nan
+  public :: ieee_is_negative
+  public :: ieee_is_normal
+  public :: ieee_nearest
+  public :: ieee_other
+  public :: ieee_next_after
+  public :: ieee_next_down
+  public :: ieee_next_up
+  public :: ieee_real
+  public :: ieee_round_type
+  public :: ieee_scalb
+  public :: ieee_selected_real_kind
+  public :: ieee_support_datatype
+  public :: ieee_support_denormal
+  public :: ieee_support_divide
+  public :: ieee_support_inf
+  public :: ieee_support_io
+  public :: ieee_support_nan
+  public :: ieee_support_rounding
+  public :: ieee_support_sqrt
+  public :: ieee_support_standard
+  public :: ieee_support_subnormal
+  public :: ieee_support_underflow_control
+  public :: ieee_to_zero
+  public :: ieee_up
+
+  ! Explicitly export the symbols from __fortran_ieee_exceptions
+  public :: ieee_flag_type
+  public :: ieee_invalid
+  public :: ieee_overflow
+  public :: ieee_divide_by_zero
+  public :: ieee_underflow
+  public :: ieee_inexact
+  public :: ieee_denorm
+  public :: ieee_usual
+  public :: ieee_all
+  public :: ieee_modes_type
+  public :: ieee_status_type
+  public :: ieee_get_flag
+  public :: ieee_get_halting_mode
+  public :: ieee_get_modes
+  public :: ieee_get_status
+  public :: ieee_set_flag
+  public :: ieee_set_halting_mode
+  public :: ieee_set_modes
+  public :: ieee_set_status
+  public :: ieee_support_flag
+  public :: ieee_support_halting
+
+  type, public :: ieee_class_type
     private
     integer(kind=1) :: which = 0
   end type ieee_class_type
 
-  type(ieee_class_type), parameter :: &
+  type(ieee_class_type), parameter, public :: &
     ieee_signaling_nan = ieee_class_type(_FORTRAN_RUNTIME_IEEE_SIGNALING_NAN), &
     ieee_quiet_nan = ieee_class_type(_FORTRAN_RUNTIME_IEEE_QUIET_NAN), &
     ieee_negative_inf = ieee_class_type(_FORTRAN_RUNTIME_IEEE_NEGATIVE_INF), &
@@ -62,23 +131,10 @@ module ieee_arithmetic
     ieee_positive_inf = ieee_class_type(_FORTRAN_RUNTIME_IEEE_POSITIVE_INF), &
     ieee_other_value = ieee_class_type(_FORTRAN_RUNTIME_IEEE_OTHER_VALUE)
 
-  type(ieee_class_type), parameter :: &
+  type(ieee_class_type), parameter, public :: &
     ieee_negative_denormal = ieee_negative_subnormal, &
     ieee_positive_denormal = ieee_positive_subnormal
 
-  type :: ieee_round_type
-    private
-    integer(kind=1) :: mode = 0
-  end type ieee_round_type
-
-  type(ieee_round_type), parameter :: &
-    ieee_to_zero = ieee_round_type(_FORTRAN_RUNTIME_IEEE_TO_ZERO), &
-    ieee_nearest = ieee_round_type(_FORTRAN_RUNTIME_IEEE_NEAREST), &
-    ieee_up = ieee_round_type(_FORTRAN_RUNTIME_IEEE_UP), &
-    ieee_down = ieee_round_type(_FORTRAN_RUNTIME_IEEE_DOWN), &
-    ieee_away = ieee_round_type(_FORTRAN_RUNTIME_IEEE_AWAY), &
-    ieee_other = ieee_round_type(_FORTRAN_RUNTIME_IEEE_OTHER)
-
   interface operator(==)
     elemental logical function ieee_class_eq(x, y)
       import ieee_class_type
@@ -89,6 +145,8 @@ module ieee_arithmetic
       type(ieee_round_type), intent(in) :: x, y
     end function ieee_round_eq
   end interface operator(==)
+  public :: operator(==)
+
   interface operator(/=)
     elemental logical function ieee_class_ne(x, y)
       import ieee_class_type
@@ -99,14 +157,21 @@ module ieee_arithmetic
       type(ieee_round_type), intent(in) :: x, y
     end function ieee_round_ne
   end interface operator(/=)
-  private :: ieee_class_eq, ieee_round_eq, ieee_class_ne, ieee_round_ne
+  public :: operator(/=)
 
 ! Define specifics with 1 or 2 INTEGER, LOGICAL, or REAL arguments for
 ! generic G.
+!
+! The result type of most function specifics is either a fixed type or
+! the type of the first argument. The result type of a SPECIFICS_rRR
+! function call is the highest precision argument type.
+
 #define SPECIFICS_I(G) \
   G(1) G(2) G(4) G(8) G(16)
 #define SPECIFICS_L(G) \
   G(1) G(2) G(4) G(8)
+
+#if FLANG_SUPPORT_R16
 #if __x86_64__
 #define SPECIFICS_R(G) \
   G(2) G(3) G(4) G(8) G(10) G(16)
@@ -114,12 +179,24 @@ module ieee_arithmetic
 #define SPECIFICS_R(G) \
   G(2) G(3) G(4) G(8) G(16)
 #endif
+#else
+#if __x86_64__
+#define SPECIFICS_R(G) \
+  G(2) G(3) G(4) G(8) G(10)
+#else
+#define SPECIFICS_R(G) \
+  G(2) G(3) G(4) G(8)
+#endif
+#endif
+
 #define SPECIFICS_II(G) \
   G(1,1) G(1,2) G(1,4) G(1,8) G(1,16) \
   G(2,1) G(2,2) G(2,4) G(2,8) G(2,16) \
   G(4,1) G(4,2) G(4,4) G(4,8) G(4,16) \
   G(8,1) G(8,2) G(8,4) G(8,8) G(8,16) \
   G(16,1) G(16,2) G(16,4) G(16,8) G(16,16)
+
+#if FLANG_SUPPORT_R16
 #if __x86_64__
 #define SPECIFICS_RI(G) \
   G(2,1) G(2,2) G(2,4) G(2,8) G(2,16) \
@@ -136,7 +213,24 @@ module ieee_arithmetic
   G(8,1) G(8,2) G(8,4) G(8,8) G(8,16) \
   G(16,1) G(16,2) G(16,4) G(16,8) G(16,16)
 #endif
+#else
+#if __x86_64__
+#define SPECIFICS_RI(G) \
+  G(2,1) G(2,2) G(2,4) G(2,8) \
+  G(3,1) G(3,2) G(3,4) G(3,8) \
+  G(4,1) G(4,2) G(4,4) G(4,8) \
+  G(8,1) G(8,2) G(8,4) G(8,8) \
+  G(10,1) G(10,2) G(10,4) G(10,8)
+#else
+#define SPECIFICS_RI(G) \
+  G(2,1) G(2,2) G(2,4) G(2,8) \
+  G(3,1) G(3,2) G(3,4) G(3,8) \
+  G(4,1) G(4,2) G(4,4) G(4,8) \
+  G(8,1) G(8,2) G(8,4) G(8,8)
+#endif
+#endif
 
+#if FLANG_SUPPORT_R16
 #if __x86_64__
 #define SPECIFICS_RR(G) \
   G(2,2) G(2,3) G(2,4) G(2,8) G(2,10) G(2,16) \
@@ -145,6 +239,13 @@ module ieee_arithmetic
   G(8,2) G(8,3) G(8,4) G(8,8) G(8,10) G(8,16) \
   G(10,2) G(10,3) G(10,4) G(10,8) G(10,10) G(10,16) \
   G(16,2) G(16,3) G(16,4) G(16,8) G(16,10) G(16,16)
+#define SPECIFICS_rRR(G) \
+  G(2,2,2) G(2,2,3) G(4,2,4) G(8,2,8) G(10,2,10) G(16,2,16) \
+  G(2,3,2) G(3,3,3) G(4,3,4) G(8,3,8) G(10,3,10) G(16,3,16) \
+  G(4,4,2) G(4,4,3) G(4,4,4) G(8,4,8) G(10,4,10) G(16,4,16) \
+  G(8,8,2) G(8,8,3) G(8,8,4) G(8,8,8) G(10,8,10) G(16,8,16) \
+  G(10,10,2) G(10,10,3) G(10,10,4) G(10,10,8) G(10,10,10) G(16,10,16) \
+  G(16,16,2) G(16,16,3) G(16,16,4) G(16,16,8) G(16,16,10) G(16,16,16)
 #else
 #define SPECIFICS_RR(G) \
   G(2,2) G(2,3) G(2,4) G(2,8) G(2,16) \
@@ -152,58 +253,39 @@ module ieee_arithmetic
   G(4,2) G(4,3) G(4,4) G(4,8) G(4,16) \
   G(8,2) G(8,3) G(8,4) G(8,8) G(8,16) \
   G(16,2) G(16,3) G(16,4) G(16,8) G(16,16)
+#define SPECIFICS_rRR(G) \
+  G(2,2,2) G(2,2,3) G(4,2,4) G(8,2,8) G(16,2,16) \
+  G(2,3,2) G(3,3,3) G(4,3,4) G(8,3,8) G(16,3,16) \
+  G(4,4,2) G(4,4,3) G(4,4,4) G(8,4,8) G(16,4,16) \
+  G(8,8,2) G(8,8,3) G(8,8,4) G(8,8,8) G(16,8,16) \
+  G(16,16,2) G(16,16,3) G(16,16,4) G(16,16,8) G(16,16,16)
 #endif
-
-! Set PRIVATE accessibility for specifics with 1 or 2 INTEGER, LOGICAL, or REAL
-! arguments for generic G.
-#define PRIVATE_I(G) private :: \
-  G##_i1, G##_i2, G##_i4, G##_i8, G##_i16
-#define PRIVATE_L(G) private :: \
-  G##_l1, G##_l2, G##_l4, G##_l8
-#if __x86_64__
-#define PRIVATE_R(G) private :: \
-  G##_a2, G##_a3, G##_a4, G##_a8, G##_a10, G##_a16
 #else
-#define PRIVATE_R(G) private :: \
-  G##_a2, G##_a3, G##_a4, G##_a8, G##_a16
+#if __x86_64__
+#define SPECIFICS_RR(G) \
+  G(2,2) G(2,3) G(2,4) G(2,8) G(2,10) \
+  G(3,2) G(3,3) G(3,4) G(3,8) G(3,10) \
+  G(4,2) G(4,3) G(4,4) G(4,8) G(4,10) \
+  G(8,2) G(8,3) G(8,4) G(8,8) G(8,10) \
+  G(10,2) G(10,3) G(10,4) G(10,8) G(10,10)
+#define SPECIFICS_rRR(G) \
+  G(2,2,2) G(2,2,3) G(4,2,4) G(8,2,8) G(10,2,10) \
+  G(2,3,2) G(3,3,3) G(4,3,4) G(8,3,8) G(10,3,10) \
+  G(4,4,2) G(4,4,3) G(4,4,4) G(8,4,8) G(10,4,10) \
+  G(8,8,2) G(8,8,3) G(8,8,4) G(8,8,8) G(10,8,10) \
+  G(10,10,2) G(10,10,3) G(10,10,4) G(10,10,8) G(10,10,10)
+#else
+#define SPECIFICS_RR(G) \
+  G(2,2) G(2,3) G(2,4) G(2,8) \
+  G(3,2) G(3,3) G(3,4) G(3,8) \
+  G(4,2) G(4,3) G(4,4) G(4,8) \
+  G(8,2) G(8,3) G(8,4) G(8,8)
+#define SPECIFICS_rRR(G) \
+  G(2,2,2) G(2,2,3) G(4,2,4) G(8,2,8) \
+  G(2,3,2) G(3,3,3) G(4,3,4) G(8,3,8) \
+  G(4,4,2) G(4,4,3) G(4,4,4) G(8,4,8) \
+  G(8,8,2) G(8,8,3) G(8,8,4) G(8,8,8)
 #endif
-#define PRIVATE_II(G) private :: \
-  G##_i1_i1, G##_i1_i2, G##_i1_i4, G##_i1_i8, G##_i1_i16, \
-  G##_i2_i1, G##_i2_i2, G##_i2_i4, G##_i2_i8, G##_i2_i16, \
-  G##_i4_i1, G##_i4_i2, G##_i4_i4, G##_i4_i8, G##_i4_i16, \
-  G##_i8_i1, G##_i8_i2, G##_i8_i4, G##_i8_i8, G##_i8_i16, \
-  G##_i16_i1, G##_i16_i2, G##_i16_i4, G##_i16_i8, G##_i16_i16
-#if __x86_64__
-#define PRIVATE_RI(G) private :: \
-  G##_a2_i1, G##_a2_i2, G##_a2_i4, G##_a2_i8, G##_a2_i16, \
-  G##_a3_i1, G##_a3_i2, G##_a3_i4, G##_a3_i8, G##_a3_i16, \
-  G##_a4_i1, G##_a4_i2, G##_a4_i4, G##_a4_i8, G##_a4_i16, \
-  G##_a8_i1, G##_a8_i2, G##_a8_i4, G##_a8_i8, G##_a8_i16, \
-  G##_a10_i1, G##_a10_i2, G##_a10_i4, G##_a10_i8, G##_a10_i16, \
-  G##_a16_i1, G##_a16_i2, G##_a16_i4, G##_a16_i8, G##_a16_i16
-#else
-#define PRIVATE_RI(G) private :: \
-  G##_a2_i1, G##_a2_i2, G##_a2_i4, G##_a2_i8, G##_a2_i16, \
-  G##_a3_i1, G##_a3_i2, G##_a3_i4, G##_a3_i8, G##_a3_i16, \
-  G##_a4_i1, G##_a4_i2, G##_a4_i4, G##_a4_i8, G##_a4_i16, \
-  G##_a8_i1, G##_a8_i2, G##_a8_i4, G##_a8_i8, G##_a8_i16, \
-  G##_a16_i1, G##_a16_i2, G##_a16_i4, G##_a16_i8, G##_a16_i16
-#endif
-#if __x86_64__
-#define PRIVATE_RR(G) private :: \
-  G##_a2_a2, G##_a2_a3, G##_a2_a4, G##_a2_a8, G##_a2_a10, G##_a2_a16, \
-  G##_a3_a2, G##_a3_a3, G##_a3_a4, G##_a3_a8, G##_a3_a10, G##_a3_a16, \
-  G##_a4_a2, G##_a4_a3, G##_a4_a4, G##_a4_a8, G##_a4_a10, G##_a4_a16, \
-  G##_a8_a2, G##_a8_a3, G##_a8_a4, G##_a8_a8, G##_a8_a10, G##_a8_a16, \
-  G##_a10_a2, G##_a10_a3, G##_a10_a4, G##_a10_a8, G##_a10_a10, G##_a10_a16, \
-  G##_a16_a2, G##_a16_a3, G##_a16_a4, G##_a16_a8, G##_a16_a10, G##_a16_a16
-#else
-#define PRIVATE_RR(G) private :: \
-  G##_a2_a2, G##_a2_a3, G##_a2_a4, G##_a2_a8, G##_a2_a16, \
-  G##_a3_a2, G##_a3_a3, G##_a3_a4, G##_a3_a8, G##_a3_a16, \
-  G##_a4_a2, G##_a4_a3, G##_a4_a4, G##_a4_a8, G##_a4_a16, \
-  G##_a8_a2, G##_a8_a3, G##_a8_a4, G##_a8_a8, G##_a8_a16, \
-  G##_a16_a2, G##_a16_a3, G##_a16_a4, G##_a16_a8, G##_a16_a16
 #endif
 
 #define IEEE_CLASS_R(XKIND) \
@@ -214,7 +296,7 @@ module ieee_arithmetic
   interface ieee_class
     SPECIFICS_R(IEEE_CLASS_R)
   end interface ieee_class
-  PRIVATE_R(IEEE_CLASS)
+  public :: ieee_class
 #undef IEEE_CLASS_R
 
 #define IEEE_COPY_SIGN_RR(XKIND, YKIND) \
@@ -225,7 +307,7 @@ module ieee_arithmetic
   interface ieee_copy_sign
     SPECIFICS_RR(IEEE_COPY_SIGN_RR)
   end interface ieee_copy_sign
-  PRIVATE_RR(IEEE_COPY_SIGN)
+  public :: ieee_copy_sign
 #undef IEEE_COPY_SIGN_RR
 
 #define IEEE_GET_ROUNDING_MODE_I(RKIND) \
@@ -241,7 +323,7 @@ module ieee_arithmetic
     end subroutine ieee_get_rounding_mode_0
     SPECIFICS_I(IEEE_GET_ROUNDING_MODE_I)
   end interface ieee_get_rounding_mode
-  PRIVATE_I(IEEE_GET_ROUNDING_MODE)
+  public :: ieee_get_rounding_mode
 #undef IEEE_GET_ROUNDING_MODE_I
 
 #define IEEE_GET_UNDERFLOW_MODE_L(GKIND) \
@@ -251,41 +333,18 @@ module ieee_arithmetic
   interface ieee_get_underflow_mode
     SPECIFICS_L(IEEE_GET_UNDERFLOW_MODE_L)
   end interface ieee_get_underflow_mode
-  PRIVATE_L(IEEE_GET_UNDERFLOW_MODE)
+  public ::  ieee_get_underflow_mode
 #undef IEEE_GET_UNDERFLOW_MODE_L
 
-! When kind argument is present, kind(result) is value(kind), not kind(kind).
-! That is not known here, so return integer(16).
-#define IEEE_INT_R(AKIND) \
-  elemental integer function ieee_int_a##AKIND(a, round); \
-    import ieee_round_type; \
-    real(AKIND), intent(in) :: a; \
-    type(ieee_round_type), intent(in) :: round; \
-  end function ieee_int_a##AKIND;
-#define IEEE_INT_RI(AKIND, KKIND) \
-  elemental integer(16) function ieee_int_a##AKIND##_i##KKIND(a, round, kind); \
-    import ieee_round_type; \
-    real(AKIND), intent(in) :: a; \
-    type(ieee_round_type), intent(in) :: round; \
-    integer(KKIND), intent(in) :: kind; \
-  end function ieee_int_a##AKIND##_i##KKIND;
-  interface ieee_int
-    SPECIFICS_R(IEEE_INT_R)
-    SPECIFICS_RI(IEEE_INT_RI)
-  end interface ieee_int
-  PRIVATE_R(IEEE_INT)
-  PRIVATE_RI(IEEE_INT)
-#undef IEEE_INT_R
-#undef IEEE_INT_RI
-
 #define IEEE_IS_FINITE_R(XKIND) \
   elemental logical function ieee_is_finite_a##XKIND(x); \
     real(XKIND), intent(in) :: x; \
+    !dir$ ignore_tkr(d) x; \
   end function ieee_is_finite_a##XKIND;
   interface ieee_is_finite
     SPECIFICS_R(IEEE_IS_FINITE_R)
   end interface ieee_is_finite
-  PRIVATE_R(IEEE_IS_FINITE)
+  public :: ieee_is_finite
 #undef IEEE_IS_FINITE_R
 
 #define IEEE_LOGB_R(XKIND) \
@@ -295,7 +354,7 @@ module ieee_arithmetic
   interface ieee_logb
     SPECIFICS_R(IEEE_LOGB_R)
   end interface ieee_logb
-  PRIVATE_R(IEEE_LOGB)
+  public :: ieee_logb
 #undef IEEE_LOGB_R
 
 #define IEEE_MAX_R(XKIND) \
@@ -305,7 +364,7 @@ module ieee_arithmetic
   interface ieee_max
     SPECIFICS_R(IEEE_MAX_R)
   end interface ieee_max
-  PRIVATE_R(IEEE_MAX)
+  public :: ieee_max
 #undef IEEE_MAX_R
 
 #define IEEE_MAX_MAG_R(XKIND) \
@@ -315,7 +374,7 @@ module ieee_arithmetic
   interface ieee_max_mag
     SPECIFICS_R(IEEE_MAX_MAG_R)
   end interface ieee_max_mag
-  PRIVATE_R(IEEE_MAX_MAG)
+  public :: ieee_max_mag
 #undef IEEE_MAX_MAG_R
 
 #define IEEE_MAX_NUM_R(XKIND) \
@@ -325,7 +384,7 @@ module ieee_arithmetic
   interface ieee_max_num
     SPECIFICS_R(IEEE_MAX_NUM_R)
   end interface ieee_max_num
-  PRIVATE_R(IEEE_MAX_NUM)
+  public :: ieee_max_num
 #undef IEEE_MAX_NUM_R
 
 #define IEEE_MAX_NUM_MAG_R(XKIND) \
@@ -335,7 +394,7 @@ module ieee_arithmetic
   interface ieee_max_num_mag
     SPECIFICS_R(IEEE_MAX_NUM_MAG_R)
   end interface ieee_max_num_mag
-  PRIVATE_R(IEEE_MAX_NUM_MAG)
+  public :: ieee_max_num_mag
 #undef IEEE_MAX_NUM_MAG_R
 
 #define IEEE_MIN_R(XKIND) \
@@ -345,7 +404,7 @@ module ieee_arithmetic
   interface ieee_min
     SPECIFICS_R(IEEE_MIN_R)
   end interface ieee_min
-  PRIVATE_R(IEEE_MIN)
+  public :: ieee_min
 #undef IEEE_MIN_R
 
 #define IEEE_MIN_MAG_R(XKIND) \
@@ -355,7 +414,7 @@ module ieee_arithmetic
   interface ieee_min_mag
     SPECIFICS_R(IEEE_MIN_MAG_R)
   end interface ieee_min_mag
-  PRIVATE_R(IEEE_MIN_MAG)
+  public :: ieee_min_mag
 #undef IEEE_MIN_MAG_R
 
 #define IEEE_MIN_NUM_R(XKIND) \
@@ -365,7 +424,7 @@ module ieee_arithmetic
   interface ieee_min_num
     SPECIFICS_R(IEEE_MIN_NUM_R)
   end interface ieee_min_num
-  PRIVATE_R(IEEE_MIN_NUM)
+  public :: ieee_min_num
 #undef IEEE_MIN_NUM_R
 
 #define IEEE_MIN_NUM_MAG_R(XKIND) \
@@ -375,7 +434,7 @@ module ieee_arithmetic
   interface ieee_min_num_mag
     SPECIFICS_R(IEEE_MIN_NUM_MAG_R)
   end interface ieee_min_num_mag
-  PRIVATE_R(IEEE_MIN_NUM_MAG)
+  public ::ieee_min_num_mag
 #undef IEEE_MIN_NUM_MAG_R
 
 #define IEEE_QUIET_EQ_R(AKIND) \
@@ -385,7 +444,7 @@ module ieee_arithmetic
   interface ieee_quiet_eq
     SPECIFICS_R(IEEE_QUIET_EQ_R)
   end interface ieee_quiet_eq
-  PRIVATE_R(IEEE_QUIET_EQ)
+  public :: ieee_quiet_eq
 #undef IEEE_QUIET_EQ_R
 
 #define IEEE_QUIET_GE_R(AKIND) \
@@ -395,7 +454,7 @@ module ieee_arithmetic
   interface ieee_quiet_ge
     SPECIFICS_R(IEEE_QUIET_GE_R)
   end interface ieee_quiet_ge
-  PRIVATE_R(IEEE_QUIET_GE)
+  public :: ieee_quiet_ge
 #undef IEEE_QUIET_GE_R
 
 #define IEEE_QUIET_GT_R(AKIND) \
@@ -405,7 +464,7 @@ module ieee_arithmetic
   interface ieee_quiet_gt
     SPECIFICS_R(IEEE_QUIET_GT_R)
   end interface ieee_quiet_gt
-  PRIVATE_R(IEEE_QUIET_GT)
+  public :: ieee_quiet_gt
 #undef IEEE_QUIET_GT_R
 
 #define IEEE_QUIET_LE_R(AKIND) \
@@ -415,7 +474,7 @@ module ieee_arithmetic
   interface ieee_quiet_le
     SPECIFICS_R(IEEE_QUIET_LE_R)
   end interface ieee_quiet_le
-  PRIVATE_R(IEEE_QUIET_LE)
+  public :: ieee_quiet_le
 #undef IEEE_QUIET_LE_R
 
 #define IEEE_QUIET_LT_R(AKIND) \
@@ -425,7 +484,7 @@ module ieee_arithmetic
   interface ieee_quiet_lt
     SPECIFICS_R(IEEE_QUIET_LT_R)
   end interface ieee_quiet_lt
-  PRIVATE_R(IEEE_QUIET_LT)
+  public :: ieee_quiet_lt
 #undef IEEE_QUIET_LT_R
 
 #define IEEE_QUIET_NE_R(AKIND) \
@@ -435,54 +494,19 @@ module ieee_arithmetic
   interface ieee_quiet_ne
     SPECIFICS_R(IEEE_QUIET_NE_R)
   end interface ieee_quiet_ne
-  PRIVATE_R(IEEE_QUIET_NE)
+  public :: ieee_quiet_ne
 #undef IEEE_QUIET_NE_R
 
-! When kind argument is present, kind(result) is value(kind), not kind(kind).
-! That is not known here, so return real(16).
-#define IEEE_REAL_I(AKIND) \
-  elemental real function ieee_real_i##AKIND(a); \
-    integer(AKIND), intent(in) :: a; \
-  end function ieee_real_i##AKIND;
-#define IEEE_REAL_R(AKIND) \
-  elemental real function ieee_real_a##AKIND(a); \
-    real(AKIND), intent(in) :: a; \
-  end function ieee_real_a##AKIND;
-#define IEEE_REAL_II(AKIND, KKIND) \
-  elemental real(16) function ieee_real_i##AKIND##_i##KKIND(a, kind); \
-    integer(AKIND), intent(in) :: a; \
-    integer(KKIND), intent(in) :: kind; \
-  end function ieee_real_i##AKIND##_i##KKIND;
-#define IEEE_REAL_RI(AKIND, KKIND) \
-  elemental real(16) function ieee_real_a##AKIND##_i##KKIND(a, kind); \
-    real(AKIND), intent(in) :: a; \
-    integer(KKIND), intent(in) :: kind; \
-  end function ieee_real_a##AKIND##_i##KKIND;
-  interface ieee_real
-    SPECIFICS_I(IEEE_REAL_I)
-    SPECIFICS_R(IEEE_REAL_R)
-    SPECIFICS_II(IEEE_REAL_II)
-    SPECIFICS_RI(IEEE_REAL_RI)
-  end interface ieee_real
-  PRIVATE_I(IEEE_REAL)
-  PRIVATE_R(IEEE_REAL)
-  PRIVATE_II(IEEE_REAL)
-  PRIVATE_RI(IEEE_REAL)
-#undef IEEE_REAL_I
-#undef IEEE_REAL_R
-#undef IEEE_REAL_II
-#undef IEEE_REAL_RI
-
-#define IEEE_REM_RR(XKIND, YKIND) \
-  elemental real(XKIND) function ieee_rem_a##XKIND##_a##YKIND(x, y); \
+#define IEEE_REM_rRR(RKIND, XKIND, YKIND) \
+  elemental real(RKIND) function ieee_rem_a##XKIND##_a##YKIND(x, y); \
     real(XKIND), intent(in) :: x; \
     real(YKIND), intent(in) :: y; \
   end function ieee_rem_a##XKIND##_a##YKIND;
   interface ieee_rem
-    SPECIFICS_RR(IEEE_REM_RR)
+    SPECIFICS_rRR(IEEE_REM_rRR)
   end interface ieee_rem
-  PRIVATE_RR(IEEE_REM)
-#undef IEEE_REM_RR
+  public :: ieee_rem
+#undef IEEE_REM_rRR
 
 #define IEEE_RINT_R(XKIND) \
   elemental real(XKIND) function ieee_rint_a##XKIND(x, round); \
@@ -493,7 +517,7 @@ module ieee_arithmetic
   interface ieee_rint
     SPECIFICS_R(IEEE_RINT_R)
   end interface ieee_rint
-  PRIVATE_R(IEEE_RINT)
+  public :: ieee_rint
 #undef IEEE_RINT_R
 
 #define IEEE_SET_ROUNDING_MODE_I(RKIND) \
@@ -509,7 +533,7 @@ module ieee_arithmetic
     end subroutine ieee_set_rounding_mode_0
     SPECIFICS_I(IEEE_SET_ROUNDING_MODE_I)
   end interface ieee_set_rounding_mode
-  PRIVATE_I(IEEE_SET_ROUNDING_MODE)
+  public :: ieee_set_rounding_mode
 #undef IEEE_SET_ROUNDING_MODE_I
 
 #define IEEE_SET_UNDERFLOW_MODE_L(GKIND) \
@@ -519,7 +543,7 @@ module ieee_arithmetic
   interface ieee_set_underflow_mode
     SPECIFICS_L(IEEE_SET_UNDERFLOW_MODE_L)
   end interface ieee_set_underflow_mode
-  PRIVATE_L(IEEE_SET_UNDERFLOW_MODE)
+  public :: ieee_set_underflow_mode
 #undef IEEE_SET_UNDERFLOW_MODE_L
 
 #define IEEE_SIGNALING_EQ_R(AKIND) \
@@ -529,7 +553,7 @@ module ieee_arithmetic
   interface ieee_signaling_eq
     SPECIFICS_R(IEEE_SIGNALING_EQ_R)
   end interface ieee_signaling_eq
-  PRIVATE_R(IEEE_SIGNALING_EQ)
+  public :: ieee_signaling_eq
 #undef IEEE_SIGNALING_EQ_R
 
 #define IEEE_SIGNALING_GE_R(AKIND) \
@@ -539,7 +563,7 @@ module ieee_arithmetic
   interface ieee_signaling_ge
     SPECIFICS_R(IEEE_SIGNALING_GE_R)
   end interface ieee_signaling_ge
-  PRIVATE_R(IEEE_SIGNALING_GE)
+  public :: ieee_signaling_ge
 #undef IEEE_SIGNALING_GE_R
 
 #define IEEE_SIGNALING_GT_R(AKIND) \
@@ -549,7 +573,7 @@ module ieee_arithmetic
   interface ieee_signaling_gt
     SPECIFICS_R(IEEE_SIGNALING_GT_R)
   end interface ieee_signaling_gt
-  PRIVATE_R(IEEE_SIGNALING_GT)
+  public :: ieee_signaling_gt
 #undef IEEE_SIGNALING_GT_R
 
 #define IEEE_SIGNALING_LE_R(AKIND) \
@@ -559,7 +583,7 @@ module ieee_arithmetic
   interface ieee_signaling_le
     SPECIFICS_R(IEEE_SIGNALING_LE_R)
   end interface ieee_signaling_le
-  PRIVATE_R(IEEE_SIGNALING_LE)
+  public :: ieee_signaling_le
 #undef IEEE_SIGNALING_LE_R
 
 #define IEEE_SIGNALING_LT_R(AKIND) \
@@ -569,7 +593,7 @@ module ieee_arithmetic
   interface ieee_signaling_lt
     SPECIFICS_R(IEEE_SIGNALING_LT_R)
   end interface ieee_signaling_lt
-  PRIVATE_R(IEEE_SIGNALING_LT)
+  public :: ieee_signaling_lt
 #undef IEEE_SIGNALING_LT_R
 
 #define IEEE_SIGNALING_NE_R(AKIND) \
@@ -579,7 +603,7 @@ module ieee_arithmetic
   interface ieee_signaling_ne
     SPECIFICS_R(IEEE_SIGNALING_NE_R)
   end interface ieee_signaling_ne
-  PRIVATE_R(IEEE_SIGNALING_NE)
+  public :: ieee_signaling_ne
 #undef IEEE_SIGNALING_NE_R
 
 #define IEEE_SIGNBIT_R(XKIND) \
@@ -589,34 +613,18 @@ module ieee_arithmetic
   interface ieee_signbit
     SPECIFICS_R(IEEE_SIGNBIT_R)
   end interface ieee_signbit
-  PRIVATE_R(IEEE_SIGNBIT)
+  public :: ieee_signbit
 #undef IEEE_SIGNBIT_R
 
-#define IEEE_SUPPORT_ROUNDING_R(XKIND) \
-  pure logical function ieee_support_rounding_a##XKIND(round_value, x); \
-    import ieee_round_type; \
-    type(ieee_round_type), intent(in) :: round_value; \
-    real(XKIND), intent(in) :: x(..); \
-  end function ieee_support_rounding_a##XKIND;
-  interface ieee_support_rounding
-    pure logical function ieee_support_rounding_0(round_value)
-      import ieee_round_type
-      type(ieee_round_type), intent(in) :: round_value
-    end function ieee_support_rounding_0
-    SPECIFICS_R(IEEE_SUPPORT_ROUNDING_R)
-  end interface ieee_support_rounding
-  PRIVATE_R(IEEE_SUPPORT_ROUNDING)
-#undef IEEE_SUPPORT_ROUNDING_R
-
 #define IEEE_UNORDERED_RR(XKIND, YKIND) \
- elemental logical function ieee_unordered_a##XKIND##_a##YKIND(x, y); \
+  elemental logical function ieee_unordered_a##XKIND##_a##YKIND(x, y); \
     real(XKIND), intent(in) :: x; \
     real(YKIND), intent(in) :: y; \
   end function ieee_unordered_a##XKIND##_a##YKIND;
   interface ieee_unordered
     SPECIFICS_RR(IEEE_UNORDERED_RR)
   end interface ieee_unordered
-  PRIVATE_RR(IEEE_UNORDERED)
+  public :: ieee_unordered
 #undef IEEE_UNORDERED_RR
 
 #define IEEE_VALUE_R(XKIND) \
@@ -628,7 +636,7 @@ module ieee_arithmetic
   interface ieee_value
     SPECIFICS_R(IEEE_VALUE_R)
   end interface ieee_value
-  PRIVATE_R(IEEE_VALUE)
+  public :: ieee_value
 #undef IEEE_VALUE_R
 
 end module ieee_arithmetic
